;; Generated package description from transient.el  -*- no-byte-compile: t -*-
(define-package "transient" "0.0.20201221.170234" "Transient commands" '((emacs "25.1")) :keywords '("bindings") :authors '(("Jonas Bernoulli" . "jonas@bernoul.li")) :maintainer '("Jonas Bernoulli" . "jonas@bernoul.li") :url "https://github.com/magit/transient")
